from __future__ import annotations
from typing import List, Dict, Optional, Any, Tuple
from pydantic import BaseModel, Field


class UserProfileExtensions(BaseModel):
    privilegeOfCommentOnUserProfile: Optional[int]
    customTitles: Optional[Tuple]


class AvatarFrame(BaseModel):
    status: Optional[int]
    ownershipStatus: Any
    version: Optional[int]
    resourceUrl: Optional[str]
    name: Optional[str]
    icon: Optional[str]
    frameType: Optional[int]
    frameId: Optional[str]


class Author(BaseModel):
    status: Optional[int]
    isNicknameVerified: Optional[bool]
    uid: Optional[str]
    level: Optional[int]
    followingStatus: Optional[int]
    accountMembershipStatus: Optional[int]
    isGlobal: Optional[bool]
    membershipStatus: Optional[int]
    reputation: Optional[int]
    role: Optional[int]
    ndcId: Optional[int]
    membersCount: Optional[int]
    nickname: Optional[str]
    icon: Optional[str]
    avatarFrame: Optional[AvatarFrame]


class UserProfile(BaseModel):
    status: Optional[int]
    moodSticker: Any
    itemsCount: Optional[int]
    consecutiveCheckInDays: Any
    uid: Optional[str]
    modifiedTime: Optional[str]
    followingStatus: Optional[int]
    onlineStatus: Optional[int]
    accountMembershipStatus: Optional[int]
    isGlobal: Optional[bool]
    avatarFrameId: Optional[str]
    reputation: Optional[int]
    postsCount: Optional[int]
    avatarFrame: Optional[AvatarFrame]
    membersCount: Optional[int]
    nickname: Optional[str]
    mediaList: Any
    icon: Optional[str]
    isNicknameVerified: Optional[bool]
    mood: Any
    level: Optional[int]
    notificationSubscriptionStatus: Optional[int]
    pushEnabled: Optional[bool]
    membershipStatus: Optional[int]
    content: Any
    joinedCount: Optional[int]
    role: Optional[int]
    commentsCount: Optional[int]
    aminoId: Optional[str]
    ndcId: Optional[int]
    createdTime: Optional[str]
    userProfileExtensions: Optional[UserProfileExtensions]
    storiesCount: Optional[int]
    blogsCount: Optional[int]


class DeviceInfo(BaseModel):
    lastClientType: Optional[int]


class AdvancedSettings(BaseModel):
    analyticsEnabled: Optional[int]


class AccountExtensions(BaseModel):
    adsFlags: Optional[int]


class Account(BaseModel):
    adsLevel: Optional[int]
    deviceInfo: Optional[DeviceInfo]
    mediaLabAdsMigrationJuly2020: Optional[bool]
    avatarFrameId: Optional[str]
    mediaLabAdsMigrationAugust2020: Optional[bool]
    adsEnabled: Optional[bool]

    username: Any
    status: Optional[int]
    uid: Optional[str]
    modifiedTime: Optional[str]
    twitterID: Any
    activation: Optional[int]
    phoneNumberActivation: Optional[int]
    emailActivation: Optional[int]
    appleID: Any
    facebookID: Any
    nickname: Optional[str]
    mediaList: Any
    googleID: Any
    icon: Optional[str]
    securityLevel: Optional[int]
    phoneNumber: Optional[str]
    membership: Any
    advancedSettings: Optional[AdvancedSettings]
    role: Optional[int]
    aminoIdEditable: Optional[bool]
    aminoId: Optional[str]
    createdTime: Optional[str]
    extensions: Optional[AccountExtensions]
    email: Optional[str]


class Login(BaseModel):
    auid: Optional[str]
    account: Optional[Account]
    secret: Optional[str]
    apiMessage: Optional[str] = Field(alias='api:message')
    sid: Optional[str]
    apiStatuscode: Optional[int] = Field(alias='api:statuscode')
    apiDuration: Optional[str] = Field(alias='api:duration')
    apiTimestamp: Optional[str] = Field(alias='api:timestamp')
    userProfile: Optional[UserProfile]


class LinkInfo(BaseModel):
    objectId: Optional[str]
    targetCode: Optional[int]
    ndcId: Optional[int]
    fullPath: Any
    shortCode: Any
    objectType: Optional[int]
    shareURLShortCode: Optional[str]
    targetCode: Optional[int]
    ndcId: Optional[int]
    fullPath: Optional[str]
    shareURLFullPath: Optional[str]


class Agent(BaseModel):
    status: Any
    ndcId: Any
    membersCount: Optional[int]
    icon: Any
    uid: Optional[str]
    followingStatus: Optional[int]
    level: Optional[int]
    nickname: Any
    isNicknameVerified: Optional[bool]
    accountMembershipStatus: Optional[int]
    membershipStatus: Optional[int]
    isGlobal: Optional[bool]
    reputation: Optional[int]
    role: Any


class Ranking(BaseModel):
    leaderboardEnabled: Optional[bool]
    rankingTable: Optional[Tuple]
    defaultLeaderboardType: Optional[int]
    leaderboardList: Optional[Tuple]
    enabled: Optional[bool]


class Influencer(BaseModel):
    maxVipNumbers: Optional[int]
    enabled: Optional[bool]
    maxVipMonthlyFee: Optional[int]
    minVipMonthlyFee: Optional[int]
    lock: Optional[bool]


class PublicChatPrivilege(BaseModel):
    minLevel: Optional[int]
    type: Optional[int]


class PublicChat(BaseModel):
    publicChatPrivilege: Optional[PublicChatPrivilege]
    enabled: Optional[bool]


class AvChat(BaseModel):
    screeningRoomEnabled: Optional[bool]
    audioEnabled: Optional[bool]
    videoEnabled: Optional[bool]
    audio2Enabled: Optional[bool]


class ModuleChat(BaseModel):
    enabled: Optional[bool]
    spamProtectionEnabled: Optional[bool]
    avChat: Optional[AvChat]
    publicChat: Optional[PublicChat]


class TopicCategories(BaseModel):
    enabled: Optional[bool]


class Featured(BaseModel):
    layout: Optional[int]
    publicChatRoomEnabled: Optional[bool]
    memberEnabled: Optional[bool]
    enabled: Optional[bool]
    lockMember: Optional[bool]
    postEnabled: Optional[bool]


class CatalogPrivilege(BaseModel):
    type: Optional[int]


class Catalog(BaseModel):
    curationEnabled: Optional[bool]
    catalogPrivilege: Optional[CatalogPrivilege]
    enabled: Optional[bool]


class UploadPrivilege(BaseModel):
    type: Optional[int]
    minLevel: Optional[int]


class AlbumManagePrivilege(BaseModel):
    type: Optional[int]
    minLevel: Optional[int]


class SharedFolder(BaseModel):
    enabled: Optional[bool]
    albumManagePrivilege: Optional[AlbumManagePrivilege]
    uploadPrivilege: Optional[UploadPrivilege]


class ExternalContent(BaseModel):
    enabled: Optional[bool]


class PollPrivilege(BaseModel):
    type: Optional[int]


class Poll(BaseModel):
    pollPrivilege: Optional[PollPrivilege]
    enabled: Optional[bool]


class LiveModePrivilege(BaseModel):
    minLevel: Optional[int]
    type: Optional[int]


class LiveMode(BaseModel):
    liveModePrivilege: Optional[LiveModePrivilege]
    enabled: Optional[bool]


class QuizPrivilege(BaseModel):
    type: Optional[int]


class Quiz(BaseModel):
    enabled: Optional[bool]
    quizPrivilege: Optional[QuizPrivilege]


class StoryPrivilege(BaseModel):
    type: Optional[int]


class Story(BaseModel):
    enabled: Optional[bool]
    storyPrivilege: Optional[StoryPrivilege]


class PublicChatRoomsPrivilege(BaseModel):
    type: Optional[int]
    minLevel: Optional[int]


class PublicChatRooms(BaseModel):
    publicChatRoomsPrivilege: Optional[PublicChatRoomsPrivilege]
    enabled: Optional[bool]


class CatalogEntryPrivilege(BaseModel):
    type: Optional[int]


class CatalogEntry(BaseModel):
    catalogEntryPrivilege: Optional[CatalogEntryPrivilege]
    enabled: Optional[bool]


class BlogPrivilege(BaseModel):
    type: Optional[int]


class TipCustomOption(BaseModel):
    value: Any
    icon: Optional[str]


class TipOptionList(BaseModel):
    value: Optional[int]
    icon: Optional[str]


class BlogExtensionsStyle(BaseModel):
    backgroundColor: Optional[str]


class BlogExtensions(BaseModel):
    style: Optional[BlogExtensionsStyle]
    fansOnly: Optional[bool]


class TipInfo(BaseModel):
    tipOptionList: Optional[Tuple[TipOptionList, ...]]
    tipMaxCoin: Optional[int]
    tippersCount: Optional[int]
    tippable: Optional[bool]
    tipMinCoin: Optional[int]
    tipCustomOption: Optional[TipCustomOption]
    tippedCoins: Optional[int]



class Blog(BaseModel):
    globalVotesCount: Optional[int]
    globalVotedValue: Optional[int]
    votedValue: Optional[int]
    keywords: Optional[str]
    strategyInfo: Optional[str]
    mediaList: Optional[Tuple]
    style: Optional[int]
    totalQuizPlayCount: Optional[int]
    title: Optional[str]
    tipInfo: Optional[TipInfo]
    contentRating: Optional[int]
    content: Optional[str]
    needHidden: Optional[bool]
    guestVotesCount: Optional[int]
    type: Optional[int]
    status: Optional[int]
    globalCommentsCount: Optional[int]
    modifiedTime: Optional[str]
    widgetDisplayInterval: Any
    totalPollVoteCount: Optional[int]
    blogId: Optional[str]
    viewCount: Optional[int]
    author: Optional[Author]
    extensions: Optional[BlogExtensions]
    votesCount: Optional[int]
    ndcId: Optional[int]
    createdTime: Optional[str]
    endTime: Any
    commentsCount: Optional[int]


class ScreeningRoomPrivilege(BaseModel):
    minLevel: Optional[int]
    type: Optional[int]


class ScreeningRoom(BaseModel):
    screeningRoomPrivilege: Optional[ScreeningRoomPrivilege]
    enabled: Optional[bool]


class QuestionPrivilege(BaseModel):
    type: Optional[int]


class Question(BaseModel):
    enabled: Optional[bool]
    questionPrivilege: Optional[QuestionPrivilege]


class WebLinkPrivilege(BaseModel):
    type: Optional[int]


class WebLink(BaseModel):
    webLinkPrivilege: Optional[WebLinkPrivilege]
    enabled: Optional[bool]


class Privilege(BaseModel):
    type: Optional[int]


class Image(BaseModel):
    enabled: Optional[bool]
    privilege: Optional[Privilege]


class PostType(BaseModel):
    image: Optional[Image]
    webLink: Optional[WebLink]
    question: Optional[Question]
    screeningRoom: Optional[ScreeningRoom]
    blog: Optional[Blog]
    catalogEntry: Optional[CatalogEntry]
    publicChatRooms: Optional[PublicChatRooms]
    story: Optional[Story]
    quiz: Optional[Quiz]
    liveMode: Optional[LiveMode]
    poll: Optional[Poll]


class Post(BaseModel):
    postType: Optional[PostType]
    enabled: Optional[bool]


class Module(BaseModel):
    post: Optional[Post]
    externalContent: Optional[ExternalContent]
    sharedFolder: Optional[SharedFolder]
    catalog: Optional[Catalog]
    featured: Optional[Featured]
    topicCategories: Optional[TopicCategories]
    chat: Optional[ModuleChat]
    influencer: Optional[Influencer]
    ranking: Optional[Ranking]


class CustomList(BaseModel):
    url: Optional[str]
    alias: Optional[str]
    id: Optional[str]


class Page(BaseModel):
    customList: Optional[Tuple[CustomList, ...]]
    defaultList: Optional[Tuple]


class Level1(BaseModel):
    id: Optional[str]


class Level2(BaseModel):
    id: Optional[str]


class LeftSidePanelNavigation(BaseModel):
    level2: Optional[Tuple[Level2, ...]]
    level1: Optional[Tuple[Level1, ...]]


class Style(BaseModel):
    iconColor: Any


class LeftSidePanel(BaseModel):
    style: Optional[Style]
    leftSidePanelNavigation: Optional[LeftSidePanelNavigation]


class HomePage(BaseModel):
    navigation: Optional[Tuple]


class Appearance(BaseModel):
    homePage: Optional[HomePage]
    leftSidePanel: Optional[LeftSidePanel]


class WelcomeMessage(BaseModel):
    text: Any
    enabled: Any


class General(BaseModel):
    premiumFeatureEnabled: Optional[bool]
    accountMembershipEnabled: Optional[bool]
    disableLiveLayerVisible: Optional[bool]
    onlyAllowOfficialTag: Optional[bool]
    welcomeMessage: Optional[WelcomeMessage]
    facebookAppIdList: Any
    videoUploadPolicy: Optional[int]
    invitePermission: Optional[int]
    disableLiveLayerActive: Optional[bool]
    joinedTopicIdList: Optional[Tuple]
    joinedBaselineCollectionIdList: Optional[Tuple]
    disableLocation: Optional[bool]
    hasPendingReviewRequest: Optional[bool]


class Configuration(BaseModel):
    general: Optional[General]
    appearance: Optional[Appearance]
    page: Optional[Page]
    module: Optional[Module]


class ThemePack(BaseModel):
    themePackRevision: Optional[int]
    themePackUrl: Optional[str]
    themeColor: Optional[str]
    themePackHash: Optional[str]


class CommunityAdvancedSettings(BaseModel):
    rankingTable: Optional[Tuple]
    leaderboardStyle: Optional[Dict]
    catalogEnabled: Optional[bool]
    joinedBaselineCollectionIdList: Optional[Tuple]
    defaultRankingTypeInLeaderboard: Optional[int]
    newsfeedPages: Optional[Tuple]
    hasPendingReviewRequest: Optional[bool]
    welcomeMessageEnabled: Any
    facebookAppIdList: Any
    frontPageLayout: Optional[int]
    pollMinFullBarVoteCount: Optional[int]
    welcomeMessageText: Any


class Community(BaseModel):
    probationStatus: Optional[int]
    keywords: Any
    themePack: Optional[ThemePack]
    listedStatus: Optional[int]
    endpoint: Optional[str]
    userAddedTopicList: Any
    templateId: Optional[int]
    joinType: Optional[int]
    searchable: Optional[bool]
    advancedSettings: Optional[CommunityAdvancedSettings]
    promotionalMediaList: Any
    ndcId: Optional[int]
    membersCount: Optional[int]
    activeInfo: Optional[Dict]
    configuration: Optional[Configuration]
    agent: Optional[Agent]
    extensions: Any
    createdTime: Optional[str]
    mediaList: Any
    name: Optional[str]
    primaryLanguage: Optional[str]
    isStandaloneAppMonetizationEnabled: Optional[bool]
    icon: Optional[str]
    isStandaloneAppDeprecated: Optional[bool]
    tagline: Optional[str]
    content: Any
    communityHeat: Optional[int]
    link: Optional[str]
    modifiedTime: Optional[str]
    status: Optional[int]


class LinkInfoExtensions(BaseModel):
    community: Optional[Community]
    linkInfo: Optional[LinkInfo]


class LinkInfoV2(BaseModel):
    path: Optional[str]
    extensions: Optional[LinkInfoExtensions]


class BaseLinkInfo(BaseModel):
    apiStatuscode: Optional[int] = Field(alias='api:statuscode')
    apiDuration: Optional[str] = Field(alias='api:duration')
    apiMessage: Optional[str] = Field(alias='api:message')
    linkInfoV2: Optional[LinkInfoV2]
    apiTimestamp: Optional[str] = Field(alias='api:timestamp')


class ChatExtensions(BaseModel):
    viewOnly: Optional[bool]
    coHost: Optional[Tuple[str, ...]]
    membersCanInvite: Optional[bool]
    language: Optional[str]
    bm: Optional[Tuple]
    lastMembersSummaryUpdateTime: Optional[int]
    fansOnly: Optional[bool]
    channelType: Optional[int]
    creatorUid: Optional[str] = Field(alias="uid")
    visibility: Optional[int]
    bannedMemberUidList: Optional[list[str]]
    announcement: Optional[str]
    pinAnnouncement: Optional[bool]
    vvChatJoinType: Optional[int]


class LastMessageSummary(BaseModel):
    uid: Optional[str]
    type: Optional[int]
    mediaType: Optional[int]
    content: Optional[str]
    messageId: Optional[str]
    createdTime: Optional[str]
    isHidden: Optional[bool]
    mediaValue: Any


class MembersSummary(BaseModel):
    status: Optional[int]
    uid: Optional[str]
    membershipStatus: Optional[int]
    role: Optional[int]
    nickname: Optional[str]
    icon: Optional[str]


class Chat(BaseModel):
    userAddedTopicList: Optional[Tuple]
    uid: Optional[str]
    membersQuota: Optional[int]
    membersSummary: Optional[Tuple[MembersSummary, ...]]
    threadId: Optional[str]
    keywords: Any
    membersCount: Optional[int]
    strategyInfo: Optional[str]
    isPinned: Optional[bool]
    title: Optional[str]
    membershipStatus: Optional[int]
    content: Optional[str]
    needHidden: Optional[bool]
    alertOption: Optional[int]
    lastReadTime: Optional[str]
    type: Optional[int]
    status: Optional[int]
    publishToGlobal: Optional[int]
    modifiedTime: Any
    lastMessageSummary: Optional[LastMessageSummary]
    condition: Optional[int]
    icon: Optional[str]
    latestActivityTime: Optional[str]
    author: Optional[Author]
    extensions: Optional[ChatExtensions]
    ndcId: Optional[int]
    createdTime: Any


class ChatBubble(BaseModel):
    id: str = Field(alias='bubbleId')
    resourceUrl: str
    config: Dict
    color: str = None
    isNew: bool
    md5: str


class Template(BaseModel):
    backgroundMedia: Tuple
    color: str
    config: Dict
    materialUrl: str
    name: str
    com_id: int = Field(alias='ndcId')
    id: str = Field(alias='templateId')


class Mention(BaseModel):
    uid: str


class Message(BaseModel):
    author: Optional[Author]
    mediaValue: Optional[str]
    threadId: Optional[str]
    mediaType: Optional[int]
    clientRefId: Optional[int]
    messageId: Optional[str]
    uid: Optional[str]
    createdTime: Optional[str]
    type: Optional[int]
    isHidden: Optional[bool]
    includedInSummary: Optional[bool]
    chatBubbleId: Optional[str]
    chatBubbleVersion: Optional[int]
    extensions: Optional[Extensions]
    ndcId: Optional[int]
    content: Optional[str]
    chatBubble: Optional[ChatBubble]


class Extensions(BaseModel):
    replyMessageId: Optional[str]
    replyMessage: Optional[Message]
    mentionedArray: Optional[Tuple[Mention, ...]]


class Paging(BaseModel):
    nextPageToken: Optional[str]
    prevPageToken: Optional[str]


class Messages(BaseModel):
    messageList: Optional[Tuple[Message, ...]]
    paging: Optional[Paging]
    apiMessage: Optional[str] = Field(alias='api:message')
    apiStatuscode: Optional[int] = Field(alias='api:statuscode')
    apiDuration: Optional[str] = Field(alias='api:duration')
    apiTimestamp: Optional[str] = Field(alias='api:timestamp')


class WalletInfo(BaseModel):
    adsEnabled: Optional[bool]
    adsFlags: Optional[int]
    adsVideoStats: Optional[int]
    businessCoinsEnabled: Optional[bool]
    totalBusinessCoins: Optional[int]
    totalBusinessCoinsFloat: Optional[float]
    totalCoins: Optional[int]
    totalCoinsFloat: Optional[float]


class Wiki(BaseModel):
    globalVotesCount: Optional[int]
    globalVotedValue: Optional[int]
    votedValue: Optional[int]
    keywords: Optional[Any]
    mediaList: Optional[Any]
    style: Optional[int]
    author: Optional[Author]
    tipInfo: Optional[TipInfo]
    contentRating: Optional[int]
    label: Optional[str]
    content: Optional[str]
    needHidden: Optional[bool]
    guestVotesCount: Optional[int]
    status: Optional[int]
    globalCommentsCount: Optional[int]
    modifiedTime: Optional[str]
    itemId: Optional[str]
    extensions: Optional[Any]
    votesCount: Optional[int]
    ndcId: Optional[int]
    createdTime: Optional[str]
    commentsCount: Optional[int]


Message.update_forward_refs()



import logging

__all__ = ['logger']

logger = logging.getLogger()

logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
fmt = logging.Formatter(fmt="%(levelname)s: %(message)s")
handler.setFormatter(fmt)
handler.setLevel(logging.INFO)
logger.addHandler(handler)



from contextlib import suppress
from copy import copy
from edamino import objects, api
from ujson import dumps, loads
from aiohttp import (
    ClientSession,
    ClientWebSocketResponse,
    WSServerHandshakeError, ContentTypeError
)
from typing import (
    Optional,
    Dict,
    Tuple,
    List,
    Literal,
    Any,
    Union
)
from time import time, timezone
from base64 import b64encode
from hashlib import sha1
from hmac import new
from binascii import hexlify
from os import urandom
from uuid import UUID

__all__ = ['Client']


def get_timestamp() -> int:
    return int(time() * 1000)


class Client:
    __slots__ = (
        'ndc_id',
        'session',
        'headers',
        'proxy'
    )

    ndc_id: str
    proxy: Optional[str]
    session: ClientSession
    headers: Dict[str, str]

    @property
    def sid(self) -> str:
        return self.headers["NDCAUTH"][4:]

    @sid.setter
    def sid(self, sid: str) -> None:
        self.headers["NDCAUTH"] = f"sid={sid}"

    @property
    def uid(self) -> str:
        return self.headers["AUID"]

    @uid.setter
    def uid(self, uid: str) -> None:
        self.headers["AUID"] = uid

    @property
    def device_id(self) -> str:
        return self.headers["NDCDEVICEID"]

    @device_id.setter
    def device_id(self, device_id: str) -> None:
        self.headers["NDCDEVICEID"] = device_id

    def __init__(self,
                 device_id: Optional[str] = None,
                 com_id: int = 0,
                 proxy: Optional[str] = None,
                 session: Optional[ClientSession] = None) -> None:
        self.proxy = proxy
        self.set_ndc(com_id)
        self.headers = {
            "NDCDEVICEID": device_id if device_id is not None else api.DEVICE_ID
        }
        self.session = session if session is not None else ClientSession(json_serialize=dumps)

    async def __aexit__(self, *args) -> None:
        await self.session.close()

    async def __aenter__(self) -> 'Client':
        return self

    def login_sid(self, sid: str, uid: str):
        self.sid = sid
        self.uid = uid

    @staticmethod
    def gen_sig(data):
        signature = b64encode(
            bytes.fromhex("32") +
            new(
                bytes.fromhex("FBF98EB3A07A9042EE5593B10CE9F3286A69D4E2"),
                data.encode("utf-8"),
                sha1
            ).digest()
        ).decode("utf-8")

        return signature

    def set_ndc(self, com_id: int) -> None:
        if com_id != 0:
            self.ndc_id = f"x{com_id}"
        else:
            self.ndc_id = "g"

    async def request(self,
                      method: Literal['POST', 'GET', 'DELETE', 'PUT'],
                      url: str,
                      json: Optional[Dict] = None,
                      full_url: bool = False,
                      data: Optional[Union[str, bytes]] = None,
                      content_type: Optional[str] = None) -> Dict:

        """
        Sending requests in amino.
        """

        headers = self.headers

        if not full_url:
            url = f"https://service.narvii.com/api/v1/{self.ndc_id}/s/{url}"
        if json is not None:
            json['timestamp'] = get_timestamp()
            data = dumps(json)
            headers['NDC-MSG-SIG'] = self.gen_sig(data)

        if content_type is not None:
            headers = copy(self.headers)
            headers['Content-Type'] = content_type

        async with self.session.request(method=method, url=url, headers=headers, data=data, proxy=self.proxy) as resp:
            try:
                response: Dict = await resp.json(loads=loads)
            except ContentTypeError:
                text = await resp.text()
                raise api.HtmlError(text)

        if resp.status != 200:
            raise api.InvalidRequest(response['api:message'], response['api:statuscode'], response)

        return response

    async def login(self, email: str, password: str) -> objects.Login:
        data = {
            "email": email,
            "v": 2,
            "secret": f"0 {password}",
            "deviceID": self.device_id,
            "clientType": 100,
            "action": "normal"
        }

        login = objects.Login(**await self.request('POST', 'auth/login', json=data))
        self.sid = login.sid
        self.uid = login.auid
        return login

    async def get_my_communities(self, start: int = 0, size: int = 25) -> Tuple[objects.Community, ...]:
        response = await self.request('GET', f'community/joined?v=1&start={start}&size={size}')
        return tuple(map(lambda community: objects.Community(**community), response['communityList']))

    async def get_info_link(self, link: str) -> objects.LinkInfoExtensions:
        base = objects.BaseLinkInfo(**await self.request('GET', f'link-resolution?q={link}'))
        return base.linkInfoV2.extensions

    async def get_user_info(self, user_id: str) -> objects.UserProfile:
        response = await self.request('GET', f'user-profile/{user_id}')
        return objects.UserProfile(**response['userProfile'])

    async def get_link_identify(self, code: str) -> Dict:
        return await self.request('GET', f'community/link-identify?q=http%3A%2F%2Faminoapps.com%2Finvite%2F{code}')

    async def join_community(self, invitation_code: Optional[str] = None) -> Dict:
        data = {}
        if invitation_code is not None:
            data["invitationId"] = await self.get_link_identify(invitation_code)
        return await self.request('POST', 'community/join', data)

    async def leave_community(self):
        return await self.request('POST', 'community/leave')

    async def upload_media(self, data: bytes, content_type: str) -> str:
        response = await self.request('POST', "media/upload", data=data, content_type=content_type)
        return response['mediaValue']

    async def download_from_link(self, link: str) -> bytes:
        async with self.session.get(link) as response:
            f = await response.read()

        if response.status != 200:
            js_resp: Dict = loads(await response.text())
            raise api.InvalidRequest(js_resp['api:message'], js_resp['api:statuscode'], js_resp)

        return f

    async def send_image(self, image: bytes, chat_id: str) -> Dict:
        data = {
            "content": None,
            "mediaType": api.MediaType.GIF_AND_IMAGE,
            "mediaUploadValueContentType": "image/jpg",
            "mediaUhqEnabled": True,
            "mediaUploadValue": b64encode(image).decode()
        }
        return await self.request("POST", f"chat/thread/{chat_id}/message", json=data)

    async def send_audio(self, audio: bytes, chat_id: str) -> Dict:
        data = {
            "content": None,
            "type": 2,
            "mediaType": api.MediaType.AUDIO,
            "mediaUploadValue": b64encode(audio).decode()
        }
        return await self.request("POST", f"chat/thread/{chat_id}/message", json=data)

    async def send_gif(self, image: bytes, chat_id: str) -> Dict:
        data = {
            "content": None,
            "mediaType": api.MediaType.GIF_AND_IMAGE,
            "mediaUploadValueContentType": "image/gif",
            "mediaUhqEnabled": True,
            "mediaUploadValue": b64encode(image).decode()
        }
        return await self.request("POST", f"chat/thread/{chat_id}/message", json=data)

    async def send_message(self,
                           chat_id: str,
                           message: Optional[str] = None,
                           message_type: int = 0,
                           ref_id: Optional[int] = None,
                           reply: Optional[str] = None,
                           mentions: Optional[List[str]] = None,
                           embed: Optional[api.Embed] = None,
                           link_snippets_list: Optional[List[api.LinkSnippet]] = None
                           ) -> objects.Message:
        if ref_id is None:
            ref_id = int(time() / 10 % 1000000000)

        if mentions is not None:
            mentions = tuple(map(lambda mention: {"uid": mention}, mentions))

        if embed is not None:
            embed = embed.dict()

        if message is not None:
            message = message.replace("<$", "‎‏").replace("$>", "‬‭")

        if link_snippets_list:
            link_snippets_list = [snippet.dict() for snippet in link_snippets_list]

        data = {
            "type": message_type,
            "content": message,
            "clientRefId": ref_id,
            "attachedObject": embed,
            "extensions": {
                "mentionedArray": mentions,
                "linkSnippetList": link_snippets_list
            },
        }
        if reply is not None:
            data["replyMessageId"] = reply

        response = await self.request("POST", f"chat/thread/{chat_id}/message", json=data)
        return objects.Message(**response['message'])

    async def get_chats(self, start: int = 0, size: int = 100) -> Tuple[objects.Chat, ...]:
        response = await self.request('GET', f'chat/thread?type=joined-me&start={start}&size={size}')
        return tuple(map(lambda chat: objects.Chat(**chat), response['threadList']))

    async def ws_connect(self) -> ClientWebSocketResponse:
        timestamp = get_timestamp()
        url = f"{self.device_id}|{timestamp}"
        headers = {
            "NDCAUTH": f"sid={self.sid}",
            "NDCDEVICEID": self.device_id,
            "NDC-MSG-SIG": self.gen_sig(url)
        }
        for i in range(4, 0, -1):
            try:
                return await self.session.ws_connect(
                    f"wss://ws{i}.narvii.com/?signbody={self.device_id}%7C{timestamp}",
                    headers=headers
                )
            except WSServerHandshakeError:
                continue

        raise api.WebSocketConnectError("Failed to connect to remote server.")

    async def receive_ws_message(self):
        timestamp = int(time())
        ws = await self.ws_connect()

        while True:
            with suppress(TypeError):
                if time() - timestamp >= 180:
                    ws = await self.ws_connect()
                    timestamp = int(time())

                yield await ws.receive_json(loads=loads)

    async def get_from_id(self, object_id: str, object_type: int = 0) -> objects.LinkInfo:
        data = {
            "objectId": object_id,
            "targetCode": 1,
            "objectType": object_type
        }

        if self.ndc_id == "g":
            url = "https://service.narvii.com/api/v1/g/s/link-resolution"
        else:
            url = f'https://service.narvii.com/api/v1/g/s-{self.ndc_id}/link-resolution'

        base = objects.BaseLinkInfo(**await self.request('POST', url, data, True))
        return base.linkInfoV2.extensions.linkInfo

    async def get_chat_info(self, chat_id) -> objects.Chat:
        response = await self.request('GET', f'chat/thread/{chat_id}')
        return objects.Chat(**response['thread'])

    async def get_chat_messages(self, chat_id: str, size: int = 25,
                                page_token: Optional[str] = None) -> objects.Messages:
        url = f'chat/thread/{chat_id}/message?v=2&pagingType=t&size={size}'
        if page_token is not None:
            url += f"&pageToken={page_token}"
        response = await self.request('GET', url)

        return objects.Messages(**response)

    async def get_chat_messages_iter(self, chat_id: str, size: int = 100):
        ost: int = size % 100
        whole: int = size // 100
        page_token: Optional[str] = None
        for i in range(whole):
            messages = await self.get_chat_messages(chat_id, size=100, page_token=page_token)

            page_token = messages.paging.nextPageToken
            yield messages.messageList

        yield (await self.get_chat_messages(chat_id, size=ost, page_token=page_token)).messageList

    async def get_chat_users(self, chat_id: str, start: int = 0, size: int = 25) -> Tuple[objects.UserProfile, ...]:
        response = await self.request(
            'GET',
            f'chat/thread/{chat_id}/member?start={start}&size={size}&type=default&cv=1.2'
        )
        return tuple(map(lambda user: objects.UserProfile(**user), response['memberList']))

    async def get_message_info(self, chat_id: str, message_id: str) -> objects.Message:
        response = await self.request('GET', f'chat/thread/{chat_id}/message/{message_id}')
        return objects.Message(**response['message'])

    async def get_blog_info(self, blog_id: str) -> objects.Blog:
        response = await self.request('GET', f'blog/{blog_id}')
        return objects.Blog(**response['blog'])

    async def get_wiki_info(self, wiki_id: str) -> objects.Wiki:
        response = await self.request('GET', f'item/{wiki_id}')
        return objects.Wiki(**response['item'])

    async def check_in(self, tz: int = -timezone // 1000) -> Dict:
        data = {"timezone": tz}
        return await self.request('POST', 'check-in', data)

    async def lottery(self, tz: int = -timezone // 1000):
        data = {
            "timezone": tz
        }
        await self.request('POST', "check-in/lottery", data)

    async def edit_profile(self,
                           nickname: Optional[str] = None,
                           content: Optional[str] = None,
                           icon: Optional[str] = None,
                           chat_request_privilege: Optional[str] = None,
                           image_list: Optional[list] = None,
                           caption_list: Optional[list] = None,
                           background_image: Optional[str] = None,
                           background_color: Optional[str] = None,
                           titles: Optional[list] = None,
                           colors: Optional[list] = None,
                           default_bubble_id: Optional[str] = None) -> Dict:
        media_list = []
        data: Dict[str, Any] = {}

        if caption_list is not None:
            for image, caption in zip(image_list, caption_list):
                media_list.append([100, image, caption])
        else:
            if image_list is not None:
                for image in image_list:
                    media_list.append([100, image, None])
        if image_list is not None or caption_list is not None:
            data["mediaList"] = media_list
        if nickname:
            data["nickname"] = nickname
        if icon:
            data["icon"] = icon
        if content:
            data["content"] = content
        if chat_request_privilege:
            data["extensions"] = {
                "privilegeOfChatInviteRequest": chat_request_privilege
            }
        if background_image:
            data["extensions"] = {
                "style": {
                    "backgroundMediaList": [[100, background_image, None, None, None]]
                }
            }
        if background_color:
            data["extensions"] = {
                "style": {
                    "backgroundColor": background_color
                }
            }
        if default_bubble_id:
            data["extensions"] = {
                "defaultBubbleId": default_bubble_id
            }
        if titles or colors:
            tlt = []
            for titles, colors in zip(titles, colors):
                tlt.append({"title": titles, "color": colors})

            data["extensions"] = {"customTitles": tlt}

        return await self.request('POST', f'user-profile/{self.uid}', data)

    async def comment_blog(self,
                           blog_id: str,
                           message: Optional[str] = None,
                           sticker_id: Optional[str] = None,
                           reply: Optional[str] = None) -> Dict:
        data = {
            "content": message,
            "stickerId": sticker_id,
            "type": 0,
            "eventSource": "PostDetailView"
        }
        if reply is not None:
            data["respondTo"] = reply
        return await self.request('POST', f'blog/{blog_id}/{"comment" if self.ndc_id != "g" else "g-comment"}', data)

    async def comment_wiki(self,
                           wiki_id: str,
                           message: Optional[str] = None,
                           sticker_id: Optional[str] = None,
                           reply: Optional[str] = None) -> Dict:
        data = {
            "content": message,
            "stickerId": sticker_id,
            "type": 0,
            "eventSource": "UserProfileView"
        }
        if reply is not None:
            data["respondTo"] = reply
        return await self.request('POST', f'item/{wiki_id}/{"comment" if self.ndc_id != "g" else "g-comment"}', data)

    async def comment_profile(self,
                              uid: str,
                              message: Optional[str] = None,
                              sticker_id: Optional[str] = None,
                              reply: Optional[str] = None) -> Dict:
        data = {
            "content": message,
            "stickerId": sticker_id,
            "type": 0,
            "eventSource": "UserProfileView"
        }
        if reply is not None:
            data["respondTo"] = reply
        return await self.request('POST', f'user-profile/{uid}/{"comment" if self.ndc_id != "g" else "g-comment"}',
                                  data)

    async def send_active_object(self,
                                 opt_in_ads_flags: int = 2147483647,
                                 tz: int = -timezone // 1000,
                                 timers: Optional[Tuple[Dict[str, int], ...]] = None
                                 ) -> Dict:
        data = {
            "userActiveTimeChunkList": timers,
            "optInAdsFlags": opt_in_ads_flags,
            "timezone": tz
        }

        return await self.request('POST', 'community/stats/user-active-time', data)

    async def like_blog(self, blog_id: str) -> Dict:
        data = {
            "value": 4,
            "eventSource": "UserProfileView"
        }
        return await self.request('POST', f"blog/{blog_id}/vote?cv=1.2", json=data)

    async def like_blogs(self, blog_ids: List[str]) -> Dict:
        data = {
            "value": 4,
            "targetIdList": blog_ids
        }
        return await self.request('POST', f"feed/vote", json=data)

    async def get_online_users(self, start: int = 0, size: int = 25) -> Tuple[objects.UserProfile, ...]:
        response = await self.request(
            "GET",
            f'live-layer?topic=ndtopic:{self.ndc_id}:online-members&start={start}&size={size}'
        )
        return tuple(map(lambda user: objects.UserProfile(**user), response["userProfileList"]))

    async def get_all_users(self,
                            users_type: Literal['recent', 'banned', 'featured', 'leaders', 'curators'] = "recent",
                            start: int = 0,
                            size: int = 25) -> Tuple[objects.UserProfile, ...]:
        response = await self.request(
            'GET',
            f'user-profile?type={users_type}&start={start}&size={size}'
        )
        return tuple(map(lambda user: objects.UserProfile(**user), response['userProfileList']))

    async def activity(self):
        pass

    async def invite_to_chat(self, uids: List[str], chat_id: str) -> Dict:
        data = {
            "uids": uids
        }
        return await self.request('POST', f'chat/thread/{chat_id}/member/invite', data)

    async def send_coins(self,
                         coins: int,
                         blog_id: Optional[str] = None,
                         chat_id: Optional[str] = None,
                         object_id: Optional[str] = None,
                         transaction_id: Optional[str] = None
                         ) -> Dict:
        url: str = ""
        if transaction_id is None:
            transaction_id = str(UUID(hexlify(urandom(16)).decode('ascii')))

        data = {
            "coins": coins,
            "tippingContext": {"transactionId": transaction_id}
        }

        if blog_id is not None:
            url = f"blog/{blog_id}/tipping"

        if chat_id is not None:
            url = f"chat/thread/{chat_id}/tipping"

        if object_id is not None:
            data["objectId"] = object_id
            data["objectType"] = 2
            url = "tipping"

        return await self.request('POST', url, json=data)

    async def subscribe(self,
                        user_id: str,
                        auto_renew: bool = False,
                        transaction_id: Optional[str] = None,
                        ) -> Dict:
        if transaction_id is None:
            transaction_id = str(UUID(hexlify(urandom(16)).decode('ascii')))
        data = {
            "paymentContext": {
                "transactionId": transaction_id,
                "isAutoRenew": auto_renew
            }
        }
        return await self.request('POST', f'influencer/{user_id}/subscribe', json=data)

    async def get_wallet_info(self) -> objects.WalletInfo:
        response = await self.request('GET', 'wallet')
        return objects.WalletInfo(**response['wallet'])

    async def join_chat(self, chat_id: str) -> Dict:
        return await self.request('POST', f'chat/thread/{chat_id}/member/{self.uid}')

    async def delete_message(self,
                             chat_id: str,
                             message_id: str,
                             as_staff: bool = False,
                             reason: Optional[str] = None) -> Dict:
        data = {
            "adminOpName": 102
        }
        if as_staff and reason:
            data["adminOpNote"] = {"content": reason}

        if not as_staff:
            return await self.request('DELETE', f'chat/thread/{chat_id}/message/{message_id}')
        else:
            return await self.request('POST', f'chat/thread/{chat_id}/message/{message_id}/admin', data)

    async def follow(self, uids: List[str]) -> Dict:
        data = {"targetUidList": uids}
        return await self.request('POST', f'user-profile/{self.uid}/joined', data)

    async def unfollow(self, uid: str) -> Dict:
        return await self.request('DELETE', f'user-profile/{self.uid}/joined/{uid}')

    async def kick_from_chat(self, chat_id: str, uid: str, allow_rejoin: bool = True) -> Dict:
        return await self.request('DELETE',
                                  f'chat/thread/{chat_id}/member/{uid}?allowRejoin={1 if allow_rejoin else 0}')

    async def get_user_blogs(self, user_id: str, start: int = 0, size: int = 25) -> Tuple[objects.Blog]:
        response = await self.request('GET', f'blog?type=user&q={user_id}&start={start}&size={size}')
        return tuple(map(lambda blog: objects.Blog(**blog), response['blogList']))

    async def pin_announcement_from_chat(self, chat_id: str, announcement: str, pin_announcement: bool = True) -> Dict:
        data = {
            "extensions": {
                "announcement": announcement,
                "pinAnnouncement": pin_announcement
            }
        }
        return await self.request('POST', f'chat/thread/{chat_id}', data)

    async def edit_chat(self,
                        chat_id: str,
                        title: Optional[str] = None,
                        icon: Optional[str] = None,
                        content: Optional[str] = None,
                        announcement: Optional[str] = None,
                        keywords: List = None,
                        pin_announcement: bool = True,
                        publish_to_global: bool = False,
                        fans_only: bool = None
                        ) -> Dict:
        data: Dict[str, Any] = {}

        if title:
            data["title"] = title
        if content:
            data["content"] = content
        if icon:
            data["icon"] = icon
        if keywords:
            data["keywords"] = keywords
        if announcement:
            data["extensions"] = {
                "announcement": announcement,
                "pinAnnouncement": pin_announcement
            }
        if fans_only:
            data["extensions"] = {"fansOnly": fans_only}

        data["publishToGlobal"] = 0 if not publish_to_global else 1

        return await self.request('POST', f'chat/thread/{chat_id}', data)

    async def set_view_only_chat(self, chat_id: str, view_only: Literal['enable', 'disable']) -> Dict:
        return await self.request('POST', f'chat/thread/{chat_id}/view-only/{view_only}')

    async def set_background_chat(self, chat_id: str, background: Optional[bytes] = None):
        data = {
            "mediaType": api.MediaType.GIF_AND_IMAGE,
            "mediaUploadValue": b64encode(background).decode(),
            "mediaUploadValueContentType": api.ContentType.IMAGE_JPG
        }
        return await self.request('POST', f'thread/{chat_id}/member/{self.uid}/background', data)

    async def set_default_background_chat(self, chat_id: str, background_number: int = 3):
        data = {
            "media": [
                100,
                f"http://static.narvii.com/default-chat-room-background/{background_number}_00.png",
                None
            ]
        }
        return await self.request('POST', f'thread/{chat_id}/member/{self.uid}/background', data)

    async def leave_chat(self, chat_id: str):
        return await self.request('DELETE', f'chat/thread/{chat_id}/member/{self.uid}')

    async def get_bubbles(self, start: int = 0, size: int = 25) -> Tuple[objects.ChatBubble, ...]:
        response = await self.request(
            'GET',
            f'chat/chat-bubble?type=all-my-bubbles&start={start}&size={size}'
        )
        return tuple(map(lambda b: objects.ChatBubble(**b), response["chatBubbleList"]))

    async def delete_bubble(self, bubble_id: str) -> Dict:
        return await self.request('DELETE', f'chat/chat-bubble/{bubble_id}')

    async def set_bubble(self, chat_id: str, bubble_id: str, apply_to_all: bool = False) -> Dict:
        data = {
            "bubbleId": bubble_id,
            "applyToAll": 1 if apply_to_all else 0,
            "threadId": chat_id,
        }
        return await self.request('POST', 'chat/thread/apply-bubble', json=data)

    async def get_templates(self):
        response = await self.request('GET', 'chat/chat-bubble/templates')
        return tuple(map(lambda t: objects.Template(**t), response['templateList']))

    async def create_bubble(self, template_id: str, config) -> objects.ChatBubble:
        response = await self.request('POST', f'chat/chat-bubble/templates/{template_id}/generate',
                                      data=config.get_zip(),
                                      content_type=api.ContentType.APPLICATION_OCTET_STREAM)
        return objects.ChatBubble(**response['chatBubble'])

    async def update_bubble(self, bubble_id: str, config):
        response = await self.request('POST', f'chat/chat-bubble/{bubble_id}', data=config.get_zip(),
                                      content_type=api.ContentType.APPLICATION_OCTET_STREAM)
        return objects.ChatBubble(**response['chatBubble'])

    async def upload_image_bubble(self, image: bytes) -> str:
        response = await self.request('POST', 'media/upload/target/chat-bubble-thumbnail',
                                      data=image, content_type=api.ContentType.IMAGE_PNG)
        return response['mediaValue']

    async def delete_invite_code(self, invite_id: str) -> Dict:
        return await self.request('DELETE', f"community/invitation/{invite_id}")

    async def delete_blog(self, blog_id: str) -> Dict:
        return await self.request('DELETE', f"blog/{blog_id}")

    async def delete_wiki(self, wiki_id: str) -> Dict:
        return await self.request('DELETE', f"item/{wiki_id}")

    async def post_blog(self,
                        title: str,
                        content: str,
                        image_list: Optional[List[str]] = None,
                        caption_list: Optional[List] = None,
                        categories_list: Optional[List] = None,
                        backgroundColor: Optional[str] = None,
                        fansOnly: bool = False,
                        extensions: Optional[Dict] = None) -> objects.Blog:
        media_list: Optional[List] = None

        if caption_list and image_list:
            media_list = [[100, image, caption] for image, caption in zip(image_list, caption_list)]

        elif image_list:
            media_list = [[100, image, None] for image in image_list]

        data = {
            "address": None,
            "content": content,
            "title": title,
            "mediaList": media_list,
            "extensions": extensions,
            "latitude": 0,
            "longitude": 0,
            "eventSource": api.SourceTypes.GLOBAL_COMPOSE
        }

        if fansOnly:
            data["extensions"] = {"fansOnly": fansOnly}
        if backgroundColor:
            data["extensions"] = {"style": {"backgroundColor": backgroundColor}}
        if categories_list:
            data["taggedBlogCategoryIdList"] = categories_list

        response = await self.request('POST', f"blog", data)
        return objects.Blog(**response['blog'])

    async def post_wiki(self,
                        title: str,
                        content: Optional[str],
                        icon: Optional[str] = None,
                        image_list: Optional[List[str]] = None,
                        keywords: Optional[str] = None,
                        backgroundColor: Optional[str] = None,
                        fansOnly: bool = False) -> objects.Wiki:
        media_list = [[100, image, None] for image in image_list] if image_list is not None else None

        data: Dict[str, Any] = {
            "label": title,
            "content": content,
            "mediaList": media_list,
            "eventSource": api.SourceTypes.GLOBAL_COMPOSE
        }

        if icon:
            data["icon"] = icon
        if keywords:
            data["keywords"] = keywords
        if fansOnly:
            data["extensions"] = {"fansOnly": fansOnly}
        if backgroundColor:
            data["extensions"] = {"style": {"backgroundColor": backgroundColor}}

        response = await self.request('POST', "item", data)
        return objects.Wiki(**response)

    async def edit_blog(self,
                        blog_id: str,
                        title: Optional[str] = None,
                        content: Optional[str] = None,
                        image_list: Optional[list] = None,
                        categories_list: Optional[list] = None,
                        background_color: Optional[str] = None,
                        fans_only: bool = False) -> Dict:
        media_list = [[100, image, None] for image in image_list]

        data: Dict[str, Any] = {
            "address": None,
            "mediaList": media_list,
            "latitude": 0,
            "longitude": 0,
            "eventSource": "PostDetailView"
        }

        if title:
            data["title"] = title
        if content:
            data["content"] = content
        if fans_only:
            data["extensions"] = {"fansOnly": fans_only}
        if background_color:
            data["extensions"] = {"style": {"backgroundColor": background_color}}
        if categories_list:
            data["taggedBlogCategoryIdList"] = categories_list

        response = await self.request('POST', f"blog/{blog_id}", data)
        return response

    async def repost_blog(self,
                          content: Optional[str] = None,
                          blogId: Optional[str] = None,
                          wikiId: Optional[str] = None) -> objects.Blog:
        if blogId:
            refObjectId, refObjectType = blogId, api.ObjectTypes.BLOG
        elif wikiId:
            refObjectId, refObjectType = wikiId, api.ObjectTypes.ITEM
        else:
            raise api.SpecifyType()

        data = {
            "content": content,
            "refObjectId": refObjectId,
            "refObjectType": refObjectType,
            "type": 2
        }

        response = await self.request('POST', "blog", data)
        return objects.Blog(**response['blog'])

    async def get_public_chats(self, start: int = 0, size: int = 25) -> Tuple[objects.Chat]:
        response = await self.request('GET', f"live-layer/public-chats?start={start}&size={size}")
        return tuple(map(lambda o: objects.Chat(**o), response["threadList"]))

    async def get_user_wikis(self, user_id: str, start: int = 0, size: int = 25) -> Tuple[objects.Wiki]:
        response = await self.request('GET', f"item?type=user-all&start={start}&size={size}&cv=1.2&uid={user_id}")
        return tuple(map(lambda o: objects.Wiki(**o), response["itemList"]))

    async def ban(self, user_id: str, reason: str, ban_type: int = None) -> Dict:
        data = {
            "reasonType": ban_type,
            "note": {
                "content": reason
            }
        }
        response = await self.request('POST', f"user-profile/{user_id}/ban", data)
        return response

    async def unban(self, user_id: str, reason: str) -> Dict:
        data = {
            "note": {
                "content": reason
            }
        }
        response = await self.request('POST', f"user-profile/{user_id}/unban", data)
        return response

    async def claim_reputation(self, chat_id: str):
        return await self.request('POST', f'chat/thread/{chat_id}/avchat-reputation')

    async def get_vc_reputation_info(self, chat_id: str):
        return await self.request('GET', f'chat/thread/{chat_id}/avchat-reputation')

    async def start_chat(self,
                         invitee_ids: List[str],
                         chat_type: int = 0,
                         content: Optional[str] = None,
                         title: Optional[str] = None,
                         is_global: bool = False,
                         publish_to_global: bool = False) -> objects.Chat:
        data = {
            "title": title,
            "type": chat_type,
            "inviteeUids": invitee_ids,
            "initialMessageContent": content,
            "publishToGlobal": 0 if publish_to_global is False else 1
        }
        if is_global is True:
            data["eventSource"] = "GlobalComposeMenu"

        response = await self.request("POST", url='chat/thread', json=data)
        return objects.Chat(**response["thread"])

    async def send_sticker(self, chat_id: str, sticker_id: str) -> objects.Message:
        data = {
            'content': None,
            'stickerId': sticker_id,
            'type': api.MessageType.STICKER
        }

        response = await self.request('POST', f'chat/thread/{chat_id}/message', data)
        return objects.Message(**response['message'])



from asyncio import TimeoutError
from pathlib import Path
from re import search
from time import time
from os import environ

from ujson import loads
from .client import Client
from .logger import logger as log
from .context import Context
from .objects import Message, UserProfile
from .api import MessageType, MediaType, WebSocketConnectError

from dotenv import load_dotenv
from asyncio import get_event_loop, AbstractEventLoop, iscoroutinefunction, Future, wait_for
from collections import namedtuple
from typing import Optional, List, Union, Tuple, Dict, Callable, Awaitable
from functools import partial
from contextlib import suppress

__all__ = ['Bot']

load_dotenv('.env')

Handler = namedtuple('Handler', ['commands', 'media_types', 'message_types', 'callback', 'description'])

HANDLERS_COMMANDS: List[Handler] = []
HANDLERS_EVENTS: List[Handler] = []
CALLBACKS: List[Callable[[Client], Awaitable[None]]] = []

ON_READY: Optional[Callable[[UserProfile], Awaitable[None]]] = None
ON_MENTION: Optional[Callable[[Context], Awaitable[None]]] = None


class TheCommandAlreadyExists(Exception):
    pass


class IsNotCoroutineFunction(Exception):
    pass


class ArgumentsNotFound(Exception):
    pass


def get_annotations(handler: Handler, words: List[str], command: str, message: str) -> List:
    args = []
    for count, (name, annotation) in enumerate([item for item in handler.callback.__annotations__.items()][1:]):
        if name == 'args':
            string = message.replace(command, '', 1)
            args.append(string)
        else:
            with suppress(IndexError):
                word = words[count]
                args.append(annotation(word))
                message = message.replace(word, '', 1)
    return args


class Bot:
    # Most of the features are taken from the amsync library :D

    __slots__ = ('email', 'password', 'prefix', 'loop', 'sid', 'uid', 'timestamp', 'ws', 'client', 'futures')

    loop: Optional[AbstractEventLoop]

    def __init__(self, email: str, password: str, prefix: str = ""):
        self.uid = None
        self.sid = None
        self.loop = None
        self.email = email
        self.password = password
        self.prefix = prefix.lower()
        self.timestamp = None
        self.ws = None
        self.futures: List[Future] = []
        self.client = None

    def get_context(self, client: Client, msg: Message, ws):
        client_context = Client(session=client.session, device_id=client.device_id, com_id=msg.ndcId)
        client_context.login_sid(self.sid, self.uid)
        context = Context(msg=msg, client=client_context, ws=ws)
        return context

    def check_cfg(self):
        email = environ.get('email')
        password = environ.get('password')

        if email != self.email:
            self.sid = None
            return

        if password == self.password:
            self.sid = environ.get('sid')
            self.uid = environ.get('uid')
        else:
            self.sid = None

        try:
            self.timestamp = int(environ.get('timestamp'))
            if time() - self.timestamp > 60 * 60 * 12:
                self.sid = None
        except TypeError:
            self.timestamp = int(time())
            self.sid = None

    def update_cfg(self):
        self.timestamp = int(time())
        string = f"sid={self.sid}\n" \
                 f"uid={self.uid}\n" \
                 f"email={self.email}\n" \
                 f"password={self.password}\n" \
                 f"timestamp={self.timestamp}"

        with open('.env', 'w') as file:
            file.write(string)

    @staticmethod
    def event(
            message_types: Optional[Union[List[int], Tuple[int, ...]]] = None,
            media_types: Optional[Union[List[int], Tuple[int, ...]]] = None):

        if not message_types:
            message_types = [MessageType.TEXT]
        if not media_types:
            media_types = [MediaType.TEXT]

        def register_handler(callback):
            global ON_READY, ON_MENTION

            name_callback = callback.__name__

            if name_callback == "on_ready":
                ON_READY = callback
                return callback

            if name_callback == "on_mention":
                ON_MENTION = callback
                return callback

            handler = Handler(
                description=None,
                media_types=tuple(media_types),
                message_types=tuple(message_types),
                callback=callback,
                commands=None
            )
            HANDLERS_EVENTS.append(handler)
            return callback

        return register_handler

    def command(self,
                commands: Union[str, List[str]],
                description: str = "Not description.",
                message_types: Optional[Union[List[int], Tuple[int, ...]]] = None,
                media_types: Optional[Union[List[int], Tuple[int, ...]]] = None,
                prefix: Optional[str] = None):

        if isinstance(commands, str):
            commands = [commands]
        if not message_types:
            message_types = [MessageType.TEXT]
        if not media_types:
            media_types = [MediaType.TEXT]

        prefix = prefix.lower() if prefix is not None else self.prefix
        commands = [f'{prefix}{command}' for command in commands]

        def register_handler(callback):
            if iscoroutinefunction(callback) is False:
                raise IsNotCoroutineFunction(callback.__name__)

            current_command_list = [item for item_list in HANDLERS_COMMANDS for item in item_list.commands]
            for command in commands:
                if command in current_command_list:
                    raise TheCommandAlreadyExists(command)

            handler = Handler(
                commands=commands,
                description=description,
                media_types=tuple(media_types),
                message_types=tuple(message_types),
                callback=callback
            )
            HANDLERS_COMMANDS.append(handler)
            return callback

        return register_handler

    async def __call__handlers(self, data: Dict):
        if data['t'] == 1000:
            msg = Message(**data['o']['chatMessage'], ndcId=data['o']['ndcId'])

            if self.futures:
                for future in self.futures:
                    future.set_result(msg)
                self.futures.clear()

            if msg.author.uid != self.uid:
                if ON_MENTION is not None:
                    uids = ()
                    with suppress(Exception):
                        uids = (u.uid for u in msg.extensions.mentionedArray)
                    with suppress(Exception):
                        if self.uid in uids or self.uid == msg.extensions.replyMessage.author.uid:
                            self.loop.create_task(ON_MENTION(self.get_context(self.client, msg, self.ws)))

                for handler in HANDLERS_EVENTS:
                    if msg.type in handler.message_types and msg.mediaType in handler.media_types:
                        context = self.get_context(self.client, msg, self.ws)
                        self.loop.create_task(handler.callback(context))

                if msg.content is not None:
                    command = msg.content.lower()

                    for handler in HANDLERS_COMMANDS:
                        is_command_list = [command.startswith(c) for c in handler.commands]
                        if any(is_command_list):
                            context = self.get_context(self.client, msg, self.ws)
                            if '-h' in msg.content:
                                await context.reply(handler.description)
                                continue
                            args = [context]
                            current_command = handler.commands[is_command_list.index(True)]
                            content = msg.content[len(current_command):]
                            words = content.split()
                            try:
                                args += get_annotations(handler, words, current_command, content)
                            except ValueError as error:
                                log.error(repr(error) + f"\nfunction: {handler.callback.__name__}")
                                continue
                            except ArgumentsNotFound as error:
                                log.info(error)
                                continue
                            if msg.type in handler.media_types and msg.mediaType in handler.media_types:
                                try:
                                    self.loop.create_task(handler.callback(*args))
                                except TypeError as e:
                                    log.error(e)
                            break

    async def __call(self) -> None:
        timestamp: int = int(time())
        self.ws = await self.client.ws_connect()

        if ON_READY:
            await ON_READY()

        if CALLBACKS:
            async def run_while_task(cal) -> None:
                while True:
                    await cal(self.client)

            for callback in CALLBACKS:
                self.loop.create_task(run_while_task(callback))

        while True:
            try:
                if int(time()) - timestamp >= 180:
                    self.ws = await self.client.ws_connect()
                    log.info('Reconnected.')
                    timestamp = int(time())

                if self.ws.closed:
                    if time() - self.timestamp > 60 * 60 * 12:
                        login = await self.client.login(self.email, self.password)
                        log.info('Login.')
                        self.sid = login.sid
                        self.uid = login.auid
                        self.update_cfg()

                    self.ws = await self.client.ws_connect()
                    log.info('Reconnected.')

                data = await self.ws.receive_json(loads=loads)
                await self.__call__handlers(data)

            except (TypeError, KeyError, AttributeError, WebSocketConnectError):
                continue

    def start(self,
              loop: Optional[AbstractEventLoop] = None,
              device_id: Optional[str] = None,
              check_updates: bool = True) -> None:

        global ON_READY
        self.check_cfg()
        self.loop = loop if loop is not None else get_event_loop()

        self.client = Client(device_id=device_id)

        try:
            if check_updates:
                response = self.loop.run_until_complete(
                    self.client.request('GET', 'https://pypi.org/pypi/ed-amino/json', full_url=True))

                version = response['info']['version']

                with open(f'{Path(__file__).parent}/__init__.py') as f:
                    __version__ = search(r'.[0-9].[0-9].[0-9].[0-9]', f.read()).group()[1:]
                if __version__ != version:
                    log.info(f'Please update to the latest version: {version}. Current version: {__version__}')

            profile: UserProfile
            if self.sid is None:
                login = self.loop.run_until_complete(self.client.login(self.email, self.password))
                self.sid = login.sid
                self.uid = login.auid
                log.info("Update config.")
                self.update_cfg()
                profile = login.userProfile
            else:
                profile = self.loop.run_until_complete(self.client.get_user_info(self.uid))

            self.client.sid = self.sid
            self.client.uid = self.uid

            if ON_READY:
                ON_READY = partial(ON_READY, profile)

            self.loop.run_until_complete(self.__call())
        except KeyboardInterrupt:
            log.info("Goodbye. ^^")
        finally:
            self.loop.run_until_complete(self.client.session.close())

    @staticmethod
    def background_task(callback: Callable[[Client], Awaitable[None]]):
        if iscoroutinefunction(callback) is False:
            raise IsNotCoroutineFunction(callback.__name__)

        CALLBACKS.append(callback)

        return callback

    async def wait_for(self, check: Callable[[Message], bool], timeout: Optional[float] = None) -> Message:
        while True:
            index = len(self.futures)
            try:
                future = self.loop.create_future()
                self.futures.append(future)
                msg = await wait_for(future, timeout=timeout)
                if check(msg):
                    return msg
            except TimeoutError:
                del self.futures[index]
                raise TimeoutError("Message not found.")



from .client import Client
from .objects import Message
from .api import Embed, LinkSnippet
from typing import (
    List,
    Optional,
    Literal, Callable
)
from contextlib import asynccontextmanager, contextmanager
from ujson import dumps
from aiohttp import ClientWebSocketResponse


class Context:
    __slots__ = ('msg', 'client', 'ws')

    def __init__(self, msg: Message, client: Client, ws: ClientWebSocketResponse) -> None:
        self.msg = msg
        self.client = client
        self.ws = ws

    async def reply(self,
                    message: Optional[str] = None,
                    message_type: int = 0,
                    ref_id: Optional[int] = None,
                    mentions: Optional[List[str]] = None,
                    embed: Optional[Embed] = None,
                    link_snippets_list: Optional[List[LinkSnippet]] = None):
        return await self.client.send_message(message=message,
                                              chat_id=self.msg.threadId,
                                              reply=self.msg.messageId,
                                              message_type=message_type,
                                              ref_id=ref_id,
                                              mentions=mentions,
                                              embed=embed,
                                              link_snippets_list=link_snippets_list)

    async def send_image(self, image: bytes):
        return await self.client.send_image(image, chat_id=self.msg.threadId)

    async def send_gif(self, gif: bytes):
        return await self.client.send_gif(gif, chat_id=self.msg.threadId)

    async def send_audio(self, audio: bytes):
        return await self.client.send_audio(audio, chat_id=self.msg.threadId)

    async def download_from_link(self, link: str):
        return await self.client.download_from_link(link)

    async def send(self,
                   message: Optional[str] = None,
                   message_type: int = 0,
                   ref_id: Optional[int] = None,
                   mentions: Optional[List[str]] = None,
                   embed: Optional[Embed] = None,
                   link_snippets_list: Optional[List[LinkSnippet]] = None,
                   reply: Optional[str] = None):
        return await self.client.send_message(message=message,
                                              chat_id=self.msg.threadId,
                                              message_type=message_type,
                                              ref_id=ref_id,
                                              mentions=mentions,
                                              embed=embed,
                                              link_snippets_list=link_snippets_list,
                                              reply=reply)

    async def get_user_info(self):
        return await self.client.get_user_info(self.msg.author.uid)

    async def invite(self, chat_id: str):
        return await self.client.invite_to_chat(uids=[self.msg.author.uid], chat_id=chat_id)

    async def follow(self):
        return await self.client.follow([self.msg.author.uid])

    async def unfollow(self):
        return await self.client.unfollow(self.msg.uid)

    async def delete_message(self, as_staff: bool = False, reason: Optional[str] = None):
        return self.client.delete_message(chat_id=self.msg.threadId,
                                          message_id=self.msg.messageId,
                                          reason=reason,
                                          as_staff=as_staff)

    async def kick(self, allow_rejoin: bool = True):
        return self.client.kick_from_chat(self.msg.threadId, self.msg.author.uid, allow_rejoin)

    async def join_community(self, code: Optional[str] = None):
        return await self.client.join_community(code)

    async def leave_community(self):
        return await self.client.leave_community()

    async def join_chat(self):
        return await self.client.join_chat(self.msg.threadId)

    async def leave_chat(self):
        return await self.client.leave_chat(self.msg.threadId)

    async def get_info_link(self, link: str):
        return await self.client.get_info_link(link)

    async def get_from_id(self, object_id: str, object_type: int = 0):
        return await self.client.get_from_id(object_id, object_type=object_type)

    async def get_user_blogs(self, start: int = 0, size: int = 25):
        return await self.client.get_user_blogs(self.msg.author.uid, start=start, size=size)

    @asynccontextmanager
    async def typing(self, chat_type: Literal[0, 1, 2] = 2):
        data = {
            "o": {
                "actions": ["Typing"],
                "target": f"ndc://x{self.msg.ndcId}/chat-thread/{self.msg.threadId}",
                "ndcId": self.msg.ndcId,
                "params": {"threadType": chat_type},
                "id": "2713213"
            },
            "t": 304
        }

        try:
            await self.ws.send_str(dumps(data))
            yield
        finally:
            data['t'] = 306
            await self.ws.send_str(dumps(data))

    @asynccontextmanager
    async def recording(self, chat_type: int = 2):
        data = {
            "o": {
                "actions": ["Recording"],
                "target": f"ndc://x{self.msg.ndcId}/chat-thread/{self.msg.threadId}",
                "ndcId": self.msg.ndcId,
                "params": {
                    "threadType": chat_type
                },
                "id": "161486614"
            },
            "t": 304
        }
        try:
            await self.ws.send_str(dumps(data))
            yield
        finally:
            data['t'] = 306
            await self.ws.send_str(dumps(data))

    async def get_chat_messages(self, size: int = 25, page_token: Optional[str] = None):
        return await self.client.get_chat_messages(self.msg.threadId, size, page_token)

    async def start_chat(self,
                         content: Optional[str] = None,
                         chat_type: int = 0,
                         is_global: bool = False,
                         publish_to_global: bool = False):
        return await self.client.start_chat(invitee_ids=[self.msg.author.uid],
                                            content=content,
                                            chat_type=chat_type,
                                            is_global=is_global,
                                            publish_to_global=publish_to_global)

    async def send_sticker(self, sticker_id: str):
        return await self.client.send_sticker(self.msg.threadId, sticker_id)

    async def get_chat_info(self):
        return await self.client.get_chat_info(self.msg.threadId)

    @contextmanager
    def set_ndc(self, ndc_id: int = 0):
        try:
            self.client.set_ndc(ndc_id)
            yield
        finally:
            self.client.set_ndc(self.msg.ndcId)

    async def get_message_info(self):
        return await self.client.get_message_info(self.msg.threadId, self.msg.messageId)



from re import findall
from typing import Dict, List, Tuple
from string import punctuation

TAB = "\x20\x20\x20\x20"

names_used = []


def clear_key(key: str) -> str:
    for p in punctuation:
        key = key.replace(p, ' ')
    string = ''.join(k[0].upper() + k[1:] for k in key.split(' '))
    return string[0].lower() + string[1:]


def get_normal_name(name: str, name_base_class: str) -> str:
    name_new_class: str = name[0].upper() + name[1:]
    while True:
        if name_new_class in names_used:
            name_new_class = name_base_class + name_new_class
        else:
            break
    names_used.append(name_new_class)
    return name_new_class


def parse_json_to_model(json_object: Dict, name_base_class: str) -> str:
    classes: str = ''
    string: str = f'class {name_base_class}(BaseModel):\n'
    for key, value in json_object.items():
        field: str = clear_key(key)
        if field == key:
            field = ''
        else:
            _s: str = field
            field = f' = Field(None, alias="{key}")'
            key = _s
        if isinstance(value, Dict):
            name_new_class: str = get_normal_name(key, name_base_class)
            string += f'{TAB}{key}: Optional[{name_new_class}]{field}\n'
            classes += parse_json_to_model(value, name_new_class)
        elif isinstance(value, List) or isinstance(value, Tuple):
            name_new_class: str = get_normal_name(key, name_base_class)
            if all(isinstance(vl, Dict) for vl in value) and len(
                    {str(val.keys()) for val in value if isinstance(val, Dict)}) == 1:
                string += f'{TAB}{key}: Optional[Tuple[{name_new_class}, ...]]{field}\n'
                classes += parse_json_to_model(value[0], name_new_class)
            else:
                string += f'{TAB}{key}: Optional[Tuple[{", ".join(type(v).__name__ for v in value)}]]{field}\n'
        else:
            string += f'{TAB}{key}: Optional[{type(value).__name__}]{field}\n'
    string += '\n\n'
    return classes + string


def parse(json_dict: Dict, name_base_class: str, add_import_string: bool = True, clear_list_named_used: bool = True,
          class_list: str = '') -> str:
    global names_used
    names_used += [name.split(' ')[1].split('(')[0].replace(':', '') for name in findall(r'class .*', class_list)]
    name_base_class = get_normal_name(name_base_class, name_base_class)
    string = ''
    if add_import_string is True:
        string += 'from pydantic import BaseModel, Field\nfrom typing import Optional, Tuple, List, Dict, Any\n\n\n'
    string += parse_json_to_model(json_dict, name_base_class)
    if clear_list_named_used is True:
        names_used.clear()
    return string.replace('list', 'List').replace('tuple', 'Tuple').replace('dict', 'Dict').replace('NoneType', 'Any')



from base64 import b64encode
from typing import Optional, Dict, Tuple, Union, TypeVar
from zipfile import ZipFile, ZIP_DEFLATED
from io import BytesIO
from ujson import dumps
from aiofile import async_open

DEVICE_ID = "3298A71AFE0EA0B7EDDAFA2337B38488E23D8060A98946A75EAE99EBE7FA2F0B4D4212428CB8B2253D"


class SourceTypes:
    USER_PROFILE: str = "UserProfileView"
    DETAIL_POST: str = "PostDetailView"
    GLOBAL_COMPOSE: str = "GlobalComposeMenu"


class ContentType:
    AUDIO_AAC: str = "audio/aac"
    IMAGE_JPG: str = "image/jpg"
    IMAGE_PNG: str = "image/png"
    APPLICATION_URL_ENCODED: str = "application/x-www-form-urlencoded"
    APPLICATION_JSON: str = 'application/json; charset=utf-8'
    APPLICATION_OCTET_STREAM: str = 'application/octet-stream'
    APPLICATION = "application/x-www-form-urlencoded"


class InvalidRequest(Exception):
    def __init__(self, message: str, status: int, json: Dict) -> None:
        super().__init__(message)
        self.message = message
        self.status = status
        self.json = json


class WebSocketConnectError(Exception):
    pass


class SpecifyType(Exception):
    pass


class HtmlError(Exception):
    pass


class Embed:
    __slots__ = (
        'object_id',
        'object_type',
        'link',
        'title',
        'content',
        'image'
    )

    def __init__(self,
                 object_id: Optional[str] = None,
                 object_type: Optional[int] = None,
                 link: Optional[str] = None,
                 title: Optional[str] = None,
                 content: Optional[str] = None,
                 image: Optional[str] = None) -> None:
        self.object_id = object_id
        self.object_type = object_type
        self.link = link
        self.title = title
        self.content = content
        self.image = image

    def dict(self) -> Dict:
        return {
            "objectId": self.object_id,
            "objectType": self.object_type,
            "link": self.link,
            "title": self.title,
            "content": self.content,
            "mediaList": [[100, self.image, None]] if self.image is not None else None
        }


class MessageType:
    TEXT: int = 0
    STRIKE: int = 1
    VOICE: int = 2
    STICKER: int = 3
    TYPE_USER_SHARE_EXURL: int = 50
    TYPE_USER_SHARE_USER: int = 51
    VOICE_CHAT_NOT_ANSWERED: int = 52
    VOICE_CHAT_NOT_CANCELLED: int = 53
    VOICE_CHAT_NOT_DECLINED: int = 54
    VIDEO_CHAT_NOT_ANSWERED: int = 55
    VIDEO_CHAT_NOT_CANCELLED: int = 56
    VIDEO_CHAT_NOT_DECLINED: int = 57
    AVATAR_CHAT_NOT_ANSWERED: int = 58
    AVATAR_CHAT_NOT_CANCELLED: int = 59
    AVATAR_CHAT_NOT_DECLINED: int = 60
    DELETE_MESSAGE: int = 100
    GROUP_MEMBER_JOIN: int = 101
    GROUP_MEMBER_LEAVE: int = 102
    CHAT_INVITE: int = 103
    CHAT_BACKGROUND_CHANGED: int = 104
    CHAT_TITLE_CHANGED: int = 105
    CHAT_ICON_CHANGED: int = 106
    VOICE_CHAT_START: int = 107
    VIDEO_CHAT_START: int = 108
    AVATAR_CHAT_START: int = 109
    VOICE_CHAT_END: int = 110
    VIDEO_CHAT_END: int = 111
    AVATAR_CHAT_END: int = 112
    CHAT_CONTENT_CHANGED: int = 113
    SCREEN_ROOM_START: int = 114
    SCREEN_ROOM_END: int = 115
    CHAT_HOST_TRANSFERED: int = 116
    TEXT_MESSAGE_FORCE_REMOVED: int = 117
    CHAT_REMOVED_MESSAGE: int = 118
    TEXT_MESSAGE_REMOVED_BY_ADMIN: int = 119
    CHAT_TIP: int = 120
    CHAT_PIN_ANNOUNCEMENT: int = 121
    VOICE_CHAT_PERMISSION_OPEN_TO_EVERYONE: int = 122
    VOICE_CHAT_PERMISSION_INVITED_AND_REQUESTED: int = 123
    VOICE_CHAT_PERMISSION_INVITE_ONLY: int = 124
    CHAT_VIEW_ONLY_ENABLED: int = 125
    CHAT_VIEW_ONLY_DISABLED: int = 126
    CHAT_UNPIN_ANNOUNCEMENT: int = 127
    CHAT_TIPPING_ENABLED: int = 128
    CHAT_TIPPING_DISABLED: int = 129
    TIMESTAMP_MESSAGE: int = 65281
    WELCOME_MESSAGE: int = 65282
    INVITE_MESSAGE: int = 65283
    ALL: Tuple[int, ...] = (
        TEXT,
        STRIKE,
        VOICE,
        STICKER,
        TYPE_USER_SHARE_EXURL,
        TYPE_USER_SHARE_USER,
        VOICE_CHAT_NOT_ANSWERED,
        VOICE_CHAT_NOT_CANCELLED,
        VOICE_CHAT_NOT_DECLINED,
        VIDEO_CHAT_NOT_ANSWERED,
        VIDEO_CHAT_NOT_CANCELLED,
        VIDEO_CHAT_NOT_DECLINED,
        AVATAR_CHAT_NOT_ANSWERED,
        AVATAR_CHAT_NOT_CANCELLED,
        AVATAR_CHAT_NOT_DECLINED,
        DELETE_MESSAGE,
        GROUP_MEMBER_JOIN,
        GROUP_MEMBER_LEAVE,
        CHAT_INVITE,
        CHAT_BACKGROUND_CHANGED,
        CHAT_TITLE_CHANGED,
        CHAT_ICON_CHANGED,
        VOICE_CHAT_START,
        VIDEO_CHAT_START,
        AVATAR_CHAT_START,
        VOICE_CHAT_END,
        VIDEO_CHAT_END,
        AVATAR_CHAT_END,
        CHAT_CONTENT_CHANGED,
        SCREEN_ROOM_START,
        SCREEN_ROOM_END,
        CHAT_HOST_TRANSFERED,
        TEXT_MESSAGE_FORCE_REMOVED,
        CHAT_REMOVED_MESSAGE,
        TEXT_MESSAGE_REMOVED_BY_ADMIN,
        CHAT_TIP,
        CHAT_PIN_ANNOUNCEMENT,
        VOICE_CHAT_PERMISSION_OPEN_TO_EVERYONE,
        VOICE_CHAT_PERMISSION_INVITED_AND_REQUESTED,
        VOICE_CHAT_PERMISSION_INVITE_ONLY,
        CHAT_VIEW_ONLY_ENABLED,
        CHAT_VIEW_ONLY_DISABLED,
        CHAT_UNPIN_ANNOUNCEMENT,
        CHAT_TIPPING_ENABLED,
        CHAT_TIPPING_DISABLED,
        TIMESTAMP_MESSAGE,
        WELCOME_MESSAGE,
        INVITE_MESSAGE
    )


class MediaType:
    TEXT: int = 0
    AUDIO: int = 110
    GIF_AND_IMAGE: int = 100
    STICKER: int = 113
    ALL: Tuple[int, int, int, int] = (TEXT, AUDIO, GIF_AND_IMAGE, STICKER)


class File:

    @staticmethod
    def load(path: str) -> bytes:
        with open(path, 'rb') as file:
            data = file.read()
        return data

    @staticmethod
    async def async_load(path: str) -> bytes:
        async with async_open(path, 'rb') as file:
            data = await file.read()

        return data


class LinkSnippet:
    __slots__ = (
        'link',
        'media_upload_value',
        'media_type',
        'media_upload_value_content_type'
    )

    def __init__(self,
                 link: str,
                 media_upload_value: bytes,
                 media_type: int = 100,
                 media_upload_value_content_type: str = "image/png"):
        self.link = link
        self.media_upload_value = b64encode(media_upload_value).decode()
        self.media_type = media_type
        self.media_upload_value_content_type = media_upload_value_content_type

    def dict(self) -> Dict:
        return {
            "link": self.link,
            "mediaType": self.media_type,
            "mediaUploadValue": self.media_upload_value,
            "mediaUploadValueContentType": self.media_upload_value_content_type
        }


Path = TypeVar("Path", bound=str)


class Slot:
    __slots__ = (
        'image',
        'x',
        'y',
        'align',
        'sticker_id',
        'path'
    )

    def __init__(self,
                 image: Union[Path, bytes],
                 align: int,
                 x: int,
                 y: int,
                 sticker_id: Optional[str] = None) -> None:
        if isinstance(image, str):
            self.image = File.load(image)
        self.x = x
        self.y = y
        self.align = align
        self.sticker_id = sticker_id
        self.path = f"a{self.align}x{self.x}y{self.y}.png"

    def dict(self):
        return {
            "y": self.y,
            "path": self.path,
            "align": self.align,
            "x": self.x,
            "stickerId": self.sticker_id
        }


class AllowedSlot:
    __slots__ = (
        'x',
        'y',
        'align'
    )
    x: int
    y: int
    align: int

    def __init__(self, x: int, y: int, align: int):
        self.x = x
        self.y = y
        self.align = align

    def dict(self):
        return {
            "y": self.y,
            "align": self.align,
            "x": self.x
        }


class ChatBubbleConfig:
    __slots__ = (
        'image',
        'template_id',
        'name',
        'cover_image_url',
        'preview_background_url',
        'color',
        'link_color',
        'slots',
        'content_insets',
        'allowed_slots',
        'zoom_point'
    )

    def __init__(self,
                 image_or_path: Union[bytes, Path],
                 name: Optional[str] = 'Custom bubble',
                 color: Optional[str] = "#ffffff",
                 link_color: Optional[str] = '#ffffff',
                 slots: Optional[Tuple[Slot, ...]] = None,
                 cover_image_url: Optional[str] = None,
                 preview_background_url: Optional[str] = None,
                 allowed_slots: Optional[Tuple[AllowedSlot, ...]] = None,
                 content_insets: Optional[Tuple[int, int, int, int]] = None,
                 zoom_point: Optional[Tuple[int, int]] = None,
                 template_id: Optional[str] = None) -> None:

        if content_insets is None:
            content_insets = (40, 65, 35, 17)

        if zoom_point is None:
            zoom_point = (58, 43)

        self.content_insets = content_insets
        self.image = image_or_path
        self.template_id = template_id
        self.name = name
        self.cover_image_url = cover_image_url
        self.preview_background_url = preview_background_url
        self.color = color
        self.link_color = link_color
        self.slots = slots
        self.content_insets = content_insets
        self.allowed_slots = allowed_slots
        self.zoom_point = zoom_point

    def get_zip(self) -> bytes:
        config = {
            "templateId": self.template_id,
            "contentInsets": self.content_insets,
            "coverImage": self.cover_image_url,
            "id": "a7ee5618-a7aa-47ed-b68d-80088a0606e6",
            "name": self.name,
            "previewBackgroundUrl": self.preview_background_url,
            "slots": (slot.dict() for slot in self.slots) if self.slots is not None else None,
            "version": 1,
            "vertexInset": 0,
            "bubbleType": 1,
            "zoomPoint": self.zoom_point,
            'authors': 'Svyt Dola#2666 & Resq#5909',
            "color": self.color,
            "linkColor": self.link_color,
            "allowedSlots": (slot.dict() for slot in self.allowed_slots) if self.allowed_slots else None,
        }
        fm = BytesIO()

        with ZipFile(fm, 'w', ZIP_DEFLATED) as zip_arc:
            if isinstance(self.image, str):
                zip_arc.write(self.image)
                config["backgroundPath"] = self.image
            else:
                zip_arc.writestr('background.png', self.image)
                config["backgroundPath"] = 'background.png'
            zip_arc.writestr('config.json', dumps(config).encode())

            if self.slots is not None:
                for slot in self.slots:
                    zip_arc.writestr(slot.path, slot.image)

        return fm.getvalue()


class ObjectTypes:
    USER: int = 0
    BLOG: int = 1
    ITEM: int = 2
    COMMENT: int = 3
    BLOG_CATEGORY: int = 4
    BLOG_CATEGORY_ITEM_TAG: int = 5
    FEATURED_ITEM: int = 6
    CHAT_MESSAGE: int = 7

    REPUTATION_LOG_ITEM: int = 10
    POLL_OPTION: int = 11
    CHAT_THREAD: int = 12
    COMMUNITY: int = 16

    IMAGE: int = 100
    MUSIC: int = 101
    VIDEO: int = 102
    YOUTUBE: int = 103
    SHARED_FOLDER: int = 106
    FOLDER_FILE: int = 109

    VOICE: int = 110
    MODERATION_TASK: int = 111
    SCREENSHOT: int = 112
    STICKER: int = 113
    STICKER_COLLECTION: int = 114
    PROP: int = 115
    CHAT_BUBBLE: int = 116
    VIDEO_FILTER: int = 117
    ORDER: int = 118
    SHARE_REQUEST: int = 119

    VV_CHAT: int = 120
    P2A: int = 121
    AMINO_VIDEO: int = 123


class ChatType:
    PUBLIC_CHAT: int = 2
    PRIVATE_CHAT: int = 1
    PM_CHAT: int = 0
